"""Molecule class for optimisation."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

import networkx as nx
import numpy as np

if TYPE_CHECKING:
    from collections import abc

    from .atom import Atom
    from .bond import Bond


@dataclass
class Molecule:
    """Molecule to optimize.

    Parameters:
        atoms:
            Atoms that define the molecule.

        bonds:
            Bonds between atoms that define the molecule.

        position_matrix:
            A ``(n, 3)`` matrix holding the position of every atom in
            the :class:`.Molecule`.

    """

    atoms: tuple[Atom, ...]
    bonds: tuple[Bond, ...]
    position_matrix: np.ndarray

    def __post_init__(self) -> None:
        """Post initialization of molecule."""
        self.atoms = tuple(self.atoms)
        self.bonds = tuple(self.bonds)
        self.position_matrix = np.array(
            self.position_matrix.T,
            dtype=np.float64,
        )

    def get_position_matrix(self) -> np.ndarray:
        """Return a matrix holding the atomic positions.

        Returns:
            The array has the shape ``(n, 3)``. Each row holds the
            x, y and z coordinates of an atom.

        """
        return np.array(self.position_matrix.T)

    def with_displacement(self, displacement: np.ndarray) -> Molecule:
        """Return a displaced clone Molecule.

        Parameters:
            displacement:
                The displacement vector to be applied.

        """
        new_position_matrix = self.position_matrix.T + displacement
        return Molecule(
            atoms=tuple(self.atoms),
            bonds=tuple(self.bonds),
            position_matrix=np.array(new_position_matrix),
        )

    def with_position_matrix(self, position_matrix: np.ndarray) -> Molecule:
        """Return clone Molecule with new position matrix.

        Parameters:
            position_matrix:
                A position matrix of the clone. The shape of the matrix
                is ``(n, 3)``.

        """
        return Molecule(
            atoms=tuple(self.atoms),
            bonds=tuple(self.bonds),
            position_matrix=np.array(position_matrix),
        )

    def write_xyz_content(self) -> list[str]:
        """Write basic `.xyz` file content of Molecule."""
        coords = self.get_position_matrix()
        content = ["0"]
        for i, atom in enumerate(self.get_atoms(), 1):
            x, y, z = (i for i in coords[atom.get_id()])
            content.append(f"{atom.get_element_string()} {x:f} {y:f} {z:f}\n")
        # Set first line to the atom_count.
        content[0] = f"{i}\n\n"

        return content

    def write_xyz_file(self, path: str) -> None:
        """Write basic `.xyz` file of Molecule to `path`.

        Connectivity is not maintained in this file type!

        """
        content = self.write_xyz_content()

        with open(path, "w") as f:
            f.write("".join(content))

    def _write_pdb_content(self) -> list[str]:
        """Write basic `.pdb` file content of Molecule."""
        content = []
        atom_counts: dict[str, int] = {}
        hetatm = "HETATM"
        alt_loc = ""
        res_name = "UNL"
        chain_id = ""
        res_seq = "1"
        i_code = ""
        occupancy = "1.00"
        temp_factor = "0.00"

        coords = self.get_position_matrix()
        # This set will be used by bonds.
        atoms = set()
        for i, atom in enumerate(self.get_atoms(), 1):
            x, y, z = (i for i in coords[atom.get_id()])
            atom_id = atom.get_id()
            atoms.add(atom_id)
            serial = atom_id + 1
            element = atom.get_element_string()
            charge = 0
            atom_counts[element] = atom_counts.get(element, 0) + 1
            name = f"{element}{atom_counts[element]}"

            content.append(
                f"{hetatm:<6}{serial:>5} {name:<4}"
                f"{alt_loc:<1}{res_name:<3} {chain_id:<1}"
                f"{res_seq:>4}{i_code:<1}   "
                f" {x:>7.3f} {y:>7.3f} {z:>7.3f}"
                f"{occupancy:>6}{temp_factor:>6}          "
                f"{element:>2}{charge:>2}\n"
            )

        conect = "CONECT"
        for bond in self.get_bonds():
            a1 = bond.get_atom1_id()
            a2 = bond.get_atom2_id()
            if a1 in atoms and a2 in atoms:
                content.append(
                    f"{conect:<6}{a1+1:>5}{a2+1:>5}               \n"
                )

        content.append("END\n")

        return content

    def write_pdb_file(self, path: str) -> None:
        """Write basic `.pdb` file of Molecule to `path`."""
        content = self._write_pdb_content()

        with open(path, "w") as f:
            f.write("".join(content))

    def get_atoms(self) -> abc.Iterable[Atom]:
        """Yield the atoms in the molecule, ordered as input."""
        yield from self.atoms

    def get_bonds(self) -> abc.Iterable[Bond]:
        """Yield the bonds in the molecule, ordered as input."""
        yield from self.bonds

    def get_num_atoms(self) -> int:
        """Return the number of atoms in the molecule."""
        return len(self.atoms)

    def get_centroid(self, atom_ids: tuple | set | None = None) -> float:
        """Return the centroid.

        Parameters:
            atom_ids: :class:`iterable` of :class:`int`, optional
                The ids of atoms which are used to calculate the
                centroid. Can be a single :class:`int`, if a single
                atom is to be used, or ``None`` if all atoms are to be
                used.

        Returns:
            The centroid of atoms specified by `atom_ids`.

        Raises:
            If `atom_ids` has a length of ``0``.

        """
        if atom_ids is None:
            atom_ids = range(len(self.atoms))  # type: ignore[assignment]
        elif not isinstance(atom_ids, (list, tuple)):
            atom_ids = tuple(atom_ids)

        if len(atom_ids) == 0:  # type: ignore[arg-type]
            msg = "atom_ids was of length 0."
            raise ValueError(msg)

        return np.divide(
            self.position_matrix[:, atom_ids].sum(axis=1),  # type: ignore[index]
            len(atom_ids),  # type: ignore[arg-type]
        )

    def get_subunits(self, bond_pair_ids: tuple) -> dict:
        """Get connected graphs based on Molecule separated by bonds.

        Parameters:
            bond_pair_ids:
                :class:`iterable` of :class:`tuple` of :class:`ints`
                Iterable of pairs of atom ids with bond between them to
                optimize.

        Returns:
            subunits:
                The subunits of `mol` split by bonds defined by
                `bond_pair_ids`. Key is subunit identifier, Value is
                :class:`iterable` of atom ids in subunit.

        """
        # Produce a graph from the molecule that does not include edges
        # where the bonds to be optimized are.
        mol_graph = nx.Graph()
        for atom in self.get_atoms():
            mol_graph.add_node(atom.get_id())

        # Add edges.
        for bond in self.get_bonds():
            pair_ids = (bond.get_atom1_id(), bond.get_atom2_id())
            if pair_ids not in bond_pair_ids:
                mol_graph.add_edge(*pair_ids)

        # Get atom ids in disconnected subgraphs.
        return dict(enumerate(nx.connected_components(mol_graph)))

    def __str__(self) -> str:
        """String representation."""
        return repr(self)

    def __repr__(self) -> str:
        """String representation."""
        return (
            f"<{self.__class__.__name__}({len(self.atoms)} atoms) "
            f"at {id(self)}>"
        )



from __future__ import annotations

import mchammer as mch
import numpy as np
import pytest


@pytest.fixture(
    params=(
        (mch.Atom(id=0, element_string="N"), 0, "N", 1.4882656711484),
        (mch.Atom(id=65, element_string="P"), 65, "P", 1.925513788),
        (mch.Atom(id=2, element_string="C"), 2, "C", 1.60775485914852),
    )
)
def atom_info(request: pytest.FixtureRequest) -> tuple:
    return request.param


@pytest.fixture(
    params=(
        (mch.Bond(id=0, atom_ids=(0, 1)), 0, 0, 1),
        (mch.Bond(id=65, atom_ids=(2, 3)), 65, 2, 3),
        (mch.Bond(id=2, atom_ids=(3, 4)), 2, 3, 4),
        (mch.Bond(id=3, atom_ids=(0, 9)), 3, 0, 9),
    )
)
def bond_info(request: pytest.FixtureRequest) -> tuple:
    return request.param


@pytest.fixture()
def atoms() -> list[mch.Atom]:
    return [
        mch.Atom(0, "C"),
        mch.Atom(1, "C"),
        mch.Atom(2, "C"),
        mch.Atom(3, "C"),
        mch.Atom(4, "C"),
        mch.Atom(5, "C"),
    ]


@pytest.fixture()
def bonds() -> list[mch.Bond]:
    return [
        mch.Bond(0, (0, 1)),
        mch.Bond(1, (0, 2)),
        mch.Bond(2, (0, 3)),
        mch.Bond(3, (3, 4)),
        mch.Bond(4, (3, 5)),
    ]


@pytest.fixture()
def position_matrix() -> np.ndarray:
    return np.array(
        [
            [0, 1, 0],
            [1, 1, 0],
            [-1, 1, 0],
            [0, 10, 0],
            [1, 10, 0],
            [-1, 10, 0],
        ]
    )


@pytest.fixture()
def position_matrix2() -> np.ndarray:
    return np.array(
        [
            [0, 1, 0],
            [1, 1, 0],
            [-1, 1, 0],
            [0, 20, 0],
            [1, 20, 0],
            [-1, 20, 0],
        ]
    )


@pytest.fixture()
def centroid() -> np.ndarray:
    return np.array([0, 5.5, 0])


@pytest.fixture()
def molecule(
    atoms: tuple,
    bonds: tuple,
    position_matrix: np.ndarray,
) -> mch.Molecule:
    return mch.Molecule(
        atoms=atoms, bonds=bonds, position_matrix=position_matrix
    )


@pytest.fixture()
def displacement() -> np.ndarray:
    return np.array([0, 1, 0])


@pytest.fixture()
def displaced_position_matrix() -> np.ndarray:
    return np.array(
        [
            [0, 2, 0],
            [1, 2, 0],
            [-1, 2, 0],
            [0, 11, 0],
            [1, 11, 0],
            [-1, 11, 0],
        ]
    )


@pytest.fixture()
def num_atoms() -> int:
    return 6


@pytest.fixture()
def bond_vector() -> np.ndarray:
    return np.array([0, 9, 0])


@pytest.fixture()
def bond_potentials() -> list[int]:
    return [50, 0, 50, 200, 450, 800, 1250]


@pytest.fixture()
def nonbond_potentials() -> list[float]:
    return [
        34.559999999999995,
        4.319999999999999,
        1.2799999999999998,
        0.5399999999999999,
        0.27647999999999995,
        0.15999999999999998,
        0.10075801749271138,
    ]


@pytest.fixture()
def nonbonded_potential() -> float:
    return 147.2965949864993


@pytest.fixture()
def system_potential() -> float:
    return 2597.2965949864993


@pytest.fixture()
def subunits() -> dict:
    return {0: {0, 1, 2}, 1: {3, 4, 5}}


@pytest.fixture()
def position_matrix3() -> np.ndarray:
    return np.array(
        [
            [0, 1, 0],
            [1, 1, 0],
            [-1, 1, 0],
            [0, 5, 0],
            [1, 5, 0],
            [-1, 5, 0],
        ]
    )


@pytest.fixture()
def optimizer() -> mch.Optimizer:
    return mch.Optimizer(step_size=0.1, target_bond_length=2.0, num_steps=100)


@pytest.fixture()
def coll_atoms() -> list[mch.Atom]:
    return [
        mch.Atom(0, "C"),
        mch.Atom(1, "C"),
        mch.Atom(2, "C"),
        mch.Atom(3, "C"),
        mch.Atom(4, "C"),
        mch.Atom(5, "C"),
    ]


@pytest.fixture()
def coll_bonds() -> list[mch.Bond]:
    return [
        mch.Bond(0, (0, 1)),
        mch.Bond(1, (0, 2)),
        mch.Bond(2, (0, 3)),
        mch.Bond(3, (3, 4)),
        mch.Bond(4, (3, 5)),
    ]


@pytest.fixture()
def coll_position_matrix() -> np.ndarray:
    return np.array(
        [
            [0, 1, 0],
            [1, 1, 0],
            [-1, 1, 0],
            [0, 10, 0],
            [1, 10, 0],
            [-1, 10, 0],
        ]
    )


@pytest.fixture()
def coll_vectors() -> dict:
    return {0: np.array([2, -1.5, 0]), 1: np.array([0, 2.5, 0])}


@pytest.fixture()
def coll_su_dists() -> list:
    return [
        9,
        9.055385138137417,
        9.055385138137417,
        9.055385138137417,
        9,
        9.219544457292887,
        9.055385138137417,
        9.219544457292887,
        9,
    ]


@pytest.fixture()
def coll_scales() -> dict:
    return {0: 0.2, 1: 1.0}


@pytest.fixture()
def coll_step() -> float:
    return 1.5


@pytest.fixture()
def coll_position_matrix2() -> np.ndarray:
    return np.array(
        [
            [-0.6, 1.45, 0],
            [0.4, 1.45, 0],
            [-1.6, 1.45, 0],
            [0, 6.25, 0],
            [1, 6.25, 0],
            [-1, 6.25, 0],
        ]
    )


@pytest.fixture()
def su_vectors() -> dict:
    return {0: np.array([0, -4.5, 0]), 1: np.array([0, 4.5, 0])}


@pytest.fixture()
def su_scales() -> dict:
    return {0: 1.0, 1: 1.0}


@pytest.fixture()
def coll_molecule(
    coll_atoms: list,
    coll_bonds: list,
    coll_position_matrix: np.ndarray,
) -> mch.Molecule:
    return mch.Molecule(
        atoms=coll_atoms,
        bonds=coll_bonds,
        position_matrix=coll_position_matrix,
    )


@pytest.fixture()
def coll_subunits() -> dict:
    return {0: {0, 1, 2}, 1: {3, 4, 5}}


@pytest.fixture()
def coll_final_position_matrix() -> np.ndarray:
    return np.array(
        [
            [0.0, 4.75262477, 0.0],
            [1.0, 4.75262477, 0.0],
            [-1.0, 4.75262477, 0.0],
            [0.0, 6.24737523, 0.0],
            [1.0, 6.24737523, 0.0],
            [-1.0, 6.24737523, 0.0],
        ]
    )


@pytest.fixture()
def collapser() -> mch.Collapser:
    return mch.Collapser(
        step_size=0.05,
        distance_threshold=1.5,
        scale_steps=True,
    )



import os  # noqa: INP001, D100

import matplotlib.pyplot as plt
import mchammer as mch
import stk

benzene = stk.BuildingBlock.init_from_file("benzene.mol")
benzene_atoms = [
    (atom.get_id(), atom.__class__.__name__) for atom in benzene.get_atoms()
]
benzene_bonds = []
for i, bond in enumerate(benzene.get_bonds()):
    b_ids = (bond.get_atom1().get_id(), bond.get_atom2().get_id())
    benzene_bonds.append((i, b_ids))

mch_mol = mch.Molecule(
    atoms=(mch.Atom(id=i[0], element_string=i[1]) for i in benzene_atoms),
    bonds=(mch.Bond(id=i[0], atom_ids=i[1]) for i in benzene_bonds),
    position_matrix=benzene.get_position_matrix(),
)

target_bond_length = 1.2
optimizer = mch.Optimizer(
    step_size=0.25,
    target_bond_length=target_bond_length,
    num_steps=100,
)
subunits = mch_mol.get_subunits(
    bond_pair_ids=((2, 3), (1, 5)),
)
# Get all steps.
mch_mol, mch_result = optimizer.get_trajectory(
    mol=mch_mol,
    bond_pair_ids=((2, 3), (1, 5)),
    subunits=subunits,
)
benzene = benzene.with_position_matrix(mch_mol.get_position_matrix())
benzene.write("benzene_opt.mol")

with open("benzene_opt.out", "w") as f:
    f.write(mch_result.get_log())

# Output trajectory as separate xyz files for visualisation.
if not os.path.exists("benzene_traj"):  # noqa: PTH110
    os.mkdir("benzene_traj")  # noqa: PTH102
for step, new_pos_mat in mch_result.get_trajectory():
    new_mol = mch_mol.with_position_matrix(new_pos_mat)
    new_mol.write_xyz_file(f"benzene_traj/traj_{step}.xyz")

# Plot properties for parameterisation.
data = {
    "steps": [],
    "max_bond_distances": [],
    "system_potentials": [],
    "nonbonded_potentials": [],
}
for step, prop in mch_result.get_steps_properties():
    data["steps"].append(step)
    data["max_bond_distances"].append(prop["max_bond_distance"])
    data["system_potentials"].append(prop["system_potential"])
    data["nonbonded_potentials"].append(prop["nonbonded_potential"])

# Show plotting from results to viauslise progress.
fig, ax = plt.subplots(figsize=(8, 5))
ax.plot(data["steps"], data["max_bond_distances"], c="k", lw=2)
# Set number of ticks for x-axis
ax.tick_params(axis="both", which="major", labelsize=16)
ax.set_xlim(0, None)
ax.set_ylim(0, None)
ax.set_xlabel("step", fontsize=16)
ax.set_ylabel("max long bond length [angstrom]", fontsize=16)
ax.axhline(y=target_bond_length, c="r", linestyle="--")
fig.tight_layout()
fig.savefig("benzene_maxd_vs_step.pdf", dpi=360, bbox_inches="tight")
plt.close()
# Plot energy vs timestep.
fig, ax = plt.subplots(figsize=(8, 5))
ax.plot(
    data["steps"],
    data["system_potentials"],
    c="k",
    lw=2,
    label="system potential",
)
ax.plot(
    data["steps"],
    data["nonbonded_potentials"],
    c="r",
    lw=2,
    label="nonbonded potential",
)
# Set number of ticks for x-axis
ax.tick_params(axis="both", which="major", labelsize=16)
ax.set_xlim(0, None)
ax.set_ylim(0, None)
ax.set_xlabel("step", fontsize=16)
ax.set_ylabel("potential", fontsize=16)
ax.legend(fontsize=16)
fig.tight_layout()
fig.savefig("benzene_pot_vs_step.pdf", dpi=360, bbox_inches="tight")
plt.close()



import os  # noqa: INP001, D100
from collections import defaultdict

import mchammer as mch
import stk


def get_long_bond_ids(mol: mch.Molecule) -> tuple:
    """Find long bonds in stk.ConstructedMolecule."""
    long_bond_ids = []
    for bond_infos in mol.get_bond_infos():
        if bond_infos.get_building_block() is None:
            ids = (
                bond_infos.get_bond().get_atom1().get_id(),
                bond_infos.get_bond().get_atom2().get_id(),
            )
            long_bond_ids.append(ids)

    return tuple(long_bond_ids)


def get_subunits(mol: mch.Molecule) -> dict:
    """Get connected graphs based on building block ids.

    Returns:
    -------
    subunits : :class:`.dict`
        The subunits of `mol` split by building block id. Key is
        subunit identifier, Value is :class:`iterable` of atom ids in
        subunit.

    """
    subunits = defaultdict(list)
    for atom_info in mol.get_atom_infos():
        subunits[atom_info.get_building_block_id()].append(
            atom_info.get_atom().get_id()
        )

    return subunits


# Building a cage from the examples on the stk docs.
bb1 = stk.BuildingBlock(
    smiles="O=CC(C=O)C=O",
    functional_groups=[stk.AldehydeFactory()],
)
bb2 = stk.BuildingBlock(
    smiles="O=CC(Cl)(C=O)C=O",
    functional_groups=[stk.AldehydeFactory()],
)
bb3 = stk.BuildingBlock("NCCN", [stk.PrimaryAminoFactory()])
bb4 = stk.BuildingBlock.init_from_file(
    "some_complex.mol",
    functional_groups=[stk.PrimaryAminoFactory()],
)
bb5 = stk.BuildingBlock("NCCCCN", [stk.PrimaryAminoFactory()])

cage = stk.ConstructedMolecule(
    topology_graph=stk.cage.FourPlusSix(
        # building_blocks is now a dict, which maps building
        # blocks to the id of the vertices it should be placed
        # on. You can use ranges to specify the ids.
        building_blocks={
            bb1: range(2),
            bb2: (2, 3),
            bb3: 4,
            bb4: 5,
            bb5: range(6, 10),
        },
    ),
)
cage.write("poc.mol")
stk_long_bond_ids = get_long_bond_ids(cage)
mch_mol = mch.Molecule(
    atoms=(
        mch.Atom(
            id=atom.get_id(),
            element_string=atom.__class__.__name__,
        )
        for atom in cage.get_atoms()
    ),
    bonds=(
        mch.Bond(
            id=i,
            atom_ids=(
                bond.get_atom1().get_id(),
                bond.get_atom2().get_id(),
            ),
        )
        for i, bond in enumerate(cage.get_bonds())
    ),
    position_matrix=cage.get_position_matrix(),
)

optimizer = mch.Collapser(
    step_size=0.05,
    distance_threshold=1.2,
    scale_steps=True,
)
subunits = get_subunits(mol=cage)
# Get all steps.
mch_mol, mch_result = optimizer.get_trajectory(
    mol=mch_mol,
    bond_pair_ids=stk_long_bond_ids,
    subunits=subunits,
)

cage = cage.with_position_matrix(mch_result.get_final_position_matrix())
cage.write("coll_poc_opt.mol")

with open("coll_opt.out", "w") as f:
    f.write(mch_result.get_log())

# Output trajectory as separate xyz files for visualisation.
if not os.path.exists("coll_poc_traj"):  # noqa: PTH110
    os.mkdir("coll_poc_traj")  # noqa: PTH102
for step, new_pos_mat in mch_result.get_trajectory():
    new_mol = mch_mol.with_position_matrix(new_pos_mat)
    new_mol.write_xyz_file(f"coll_poc_traj/traj_{step}.xyz")



from collections import abc

import mchammer as mch
import numpy as np


def test_c_get_subunit_distances(
    collapser: mch.Collapser,
    coll_molecule: mch.Molecule,
    coll_subunits: dict,
    coll_su_dists: list,
    coll_position_matrix: np.ndarray,
) -> None:
    test = coll_molecule.with_position_matrix(coll_position_matrix)
    for i, dist in enumerate(
        collapser._get_subunit_distances(  # noqa: SLF001
            mol=test,
            subunits=coll_subunits,
        )
    ):
        print(i)
        assert np.isclose(dist, coll_su_dists[i], atol=1e-6)


def test_c_get_new_position_matrix(
    collapser: mch.Collapser,
    coll_position_matrix: np.ndarray,
    coll_position_matrix2: np.ndarray,
    coll_molecule: mch.Molecule,
    coll_step: float,
    coll_vectors: dict,
    coll_scales: dict,
    coll_subunits: dict,
) -> None:
    test_mol = coll_molecule.with_position_matrix(coll_position_matrix)
    test_pos_mat = collapser._get_new_position_matrix(  # noqa: SLF001
        mol=test_mol,
        subunits=coll_subunits,
        step_size=coll_step,
        vectors=coll_vectors,
        scales=coll_scales,
    )

    for i in range(len(coll_position_matrix2)):
        test = test_pos_mat[i]
        given = coll_position_matrix2[i]
        assert np.all(np.isclose(test, given, atol=1e-6))


def test_c_get_bb_vectors(
    collapser: mch.Collapser,
    coll_molecule: mch.Molecule,
    coll_subunits: dict,
    su_vectors: dict,
    su_scales: dict,
) -> None:
    test_vectors, test_scales = collapser._get_subunit_vectors(  # noqa: SLF001
        coll_molecule, coll_subunits
    )
    for t, s in zip(test_vectors, su_vectors):
        test = test_vectors[t]
        su_v = su_vectors[s]
        assert np.all(np.equal(test, su_v))
        test = test_scales[t]
        su_s = su_scales[s]
        assert np.all(np.equal(test, su_s))


def test_c_get_result(
    collapser: abc.Callable,
    coll_molecule: mch.Molecule,
    coll_final_position_matrix: np.ndarray,
) -> None:
    original_pos_mat = coll_molecule.get_position_matrix()
    subunits = coll_molecule.get_subunits(bond_pair_ids=((0, 3),))
    test_mol, results = collapser.get_result(
        mol=coll_molecule,
        bond_pair_ids=((0, 3),),
        subunits=subunits,
    )

    final_min_distance = min(
        dist
        for dist in collapser._get_subunit_distances(  # noqa: SLF001
            test_mol, subunits
        )
    )
    # Give it some wiggle room.
    assert np.isclose(1.4947505, final_min_distance, atol=1e-8)

    # Check position matrix because this code is deterministic.
    test_pos_mat = test_mol.get_position_matrix()
    for i in range(len(coll_final_position_matrix)):
        test = test_pos_mat[i]
        given = coll_final_position_matrix[i]
        assert np.all(np.isclose(test, given, atol=1e-8))

    # Test all other bond lengths are equivalent.
    for bond in coll_molecule.get_bonds():
        if (bond.get_atom1_id(), bond.get_atom2_id()) != (0, 3):
            test = mch.get_atom_distance(
                position_matrix=results.position_matrix,
                atom1_id=bond.get_atom1_id(),
                atom2_id=bond.get_atom2_id(),
            )
            bond_length = mch.get_atom_distance(
                position_matrix=original_pos_mat,
                atom1_id=bond.get_atom1_id(),
                atom2_id=bond.get_atom2_id(),
            )
            assert np.isclose(test, bond_length, atol=1e-6)


def test_c_get_trajectory(
    collapser: abc.Callable,
    coll_molecule: mch.Molecule,
    coll_final_position_matrix: np.ndarray,
) -> None:
    original_pos_mat = coll_molecule.get_position_matrix()
    subunits = coll_molecule.get_subunits(bond_pair_ids=((0, 3),))
    test_mol, results = collapser.get_trajectory(
        mol=coll_molecule,
        bond_pair_ids=((0, 3),),
        subunits=subunits,
    )
    test_mol = coll_molecule.with_position_matrix(
        results.get_final_position_matrix()
    )
    assert results.get_step_count() == 35

    final_min_distance = min(
        dist
        for dist in collapser._get_subunit_distances(  # noqa: SLF001
            test_mol, subunits
        )
    )
    # Give it some wiggle room.
    assert np.isclose(1.4947505, final_min_distance, atol=1e-8)

    # Check position matrix because this code is deterministic.
    test_pos_mat = test_mol.get_position_matrix()
    for i in range(len(coll_final_position_matrix)):
        test = test_pos_mat[i]
        given = coll_final_position_matrix[i]
        assert np.all(np.isclose(test, given, atol=1e-8))

    # Test all other bond lengths are equivalent.
    for bond in coll_molecule.get_bonds():
        if (bond.get_atom1_id(), bond.get_atom2_id()) != (0, 3):
            test = mch.get_atom_distance(
                position_matrix=results.get_final_position_matrix(),
                atom1_id=bond.get_atom1_id(),
                atom2_id=bond.get_atom2_id(),
            )
            bond_length = mch.get_atom_distance(
                position_matrix=original_pos_mat,
                atom1_id=bond.get_atom1_id(),
                atom2_id=bond.get_atom2_id(),
            )
            assert np.isclose(test, bond_length, atol=1e-6)



from __future__ import annotations

import mchammer as mch
import numpy as np


def test_opt_get_bond_vector(
    bond_vector: np.ndarray,
    position_matrix: np.ndarray,
) -> None:
    assert np.all(
        np.equal(
            bond_vector,
            mch.get_bond_vector(position_matrix, (0, 3)),
        )
    )


def test_opt_bond_potential(
    optimizer: mch.Optimizer,
    bond_potentials: list[int],
) -> None:
    for i, d in enumerate([1, 2, 3, 4, 5, 6, 7]):
        test = optimizer._bond_potential(distance=d)  # noqa: SLF001
        assert np.isclose(test, bond_potentials[i], atol=1e-5)


def test_opt_nonbond_potential(
    optimizer: mch.Optimizer,
    nonbond_potentials: list[float],
) -> None:
    for i, d in enumerate([1, 2, 3, 4, 5, 6, 7]):
        test = optimizer._nonbond_potential(distance=d)  # noqa: SLF001
        assert np.isclose(test, nonbond_potentials[i], atol=1e-5)


def test_opt_compute_nonbonded_potential(
    optimizer: mch.Optimizer,
    position_matrix: np.ndarray,
    nonbonded_potential: float,
) -> None:
    test = optimizer._compute_nonbonded_potential(  # noqa: SLF001
        position_matrix
    )
    assert test == nonbonded_potential


def test_opt_compute_potential(
    optimizer: mch.Optimizer,
    molecule: mch.Molecule,
    nonbonded_potential: float,
    system_potential: float,
) -> None:
    (
        test_system_potential,
        test_nonbond_potential,
    ) = optimizer.compute_potential(
        molecule,
        bond_pair_ids=((0, 3),),
    )
    assert test_nonbond_potential == nonbonded_potential
    assert test_system_potential == system_potential


def test_opt_translate_atoms_along_vector(
    molecule: mch.Molecule,
    position_matrix: np.ndarray,
    position_matrix3: np.ndarray,
) -> None:
    molecule = molecule.with_position_matrix(position_matrix)
    new_molecule = mch.translate_atoms_along_vector(
        mol=molecule, atom_ids=(3, 4, 5), vector=np.array([0, 5, 0])
    )
    assert np.all(
        np.equal(
            position_matrix3,
            new_molecule.get_position_matrix(),
        )
    )
    new_molecule = mch.translate_atoms_along_vector(
        mol=new_molecule, atom_ids=(3, 4, 5), vector=np.array([0, -5, 0])
    )
    print(position_matrix, new_molecule.get_position_matrix())
    assert np.all(
        np.equal(
            position_matrix,
            new_molecule.get_position_matrix(),
        )
    )


def test_opt_test_move() -> None:
    assert mch.test_move(
        beta=2,
        curr_pot=-1,
        new_pot=-2,
        generator=np.random.default_rng(),
    )


def test_opt_get_result(
    optimizer: mch.Optimizer,
    molecule: mch.Molecule,
) -> None:
    original_pos_mat = molecule.get_position_matrix()
    subunits = molecule.get_subunits(bond_pair_ids=((0, 3),))
    mol, results = optimizer.get_result(
        mol=molecule,
        bond_pair_ids=((0, 3),),
        subunits=subunits,
    )

    final_bond_length = np.linalg.norm(
        mch.get_bond_vector(
            position_matrix=results.position_matrix,
            bond_pair=(0, 3),
        ),
    )
    # Give it some wiggle room.
    assert final_bond_length > 1.5
    assert final_bond_length < 2.5

    # Test all other bond lengths are equivalent.
    for bond in molecule.get_bonds():
        if (bond.get_atom1_id(), bond.get_atom2_id()) != (0, 3):
            test = mch.get_atom_distance(
                position_matrix=results.position_matrix,
                atom1_id=bond.get_atom1_id(),
                atom2_id=bond.get_atom2_id(),
            )
            bond_length = mch.get_atom_distance(
                position_matrix=original_pos_mat,
                atom1_id=bond.get_atom1_id(),
                atom2_id=bond.get_atom2_id(),
            )
            assert np.isclose(test, bond_length, atol=1e-6)


def test_opt_get_trajectory(
    optimizer: mch.Optimizer,
    molecule: mch.Molecule,
) -> None:
    original_pos_mat = molecule.get_position_matrix()
    subunits = molecule.get_subunits(bond_pair_ids=((0, 3),))
    mol, results = optimizer.get_trajectory(
        mol=molecule,
        bond_pair_ids=((0, 3),),
        subunits=subunits,
    )

    assert results.get_step_count() == 99
    assert len(tuple(results.get_steps_properties())) == 100

    final_bond_length = np.linalg.norm(
        mch.get_bond_vector(
            position_matrix=results.get_final_position_matrix(),
            bond_pair=(0, 3),
        ),
    )
    # Give it some wiggle room.
    assert final_bond_length > 1.5
    assert final_bond_length < 2.5

    # Test all other bond lengths are equivalent.
    for bond in molecule.get_bonds():
        if (bond.get_atom1_id(), bond.get_atom2_id()) != (0, 3):
            test = mch.get_atom_distance(
                position_matrix=results.get_final_position_matrix(),
                atom1_id=bond.get_atom1_id(),
                atom2_id=bond.get_atom2_id(),
            )
            bond_length = mch.get_atom_distance(
                position_matrix=original_pos_mat,
                atom1_id=bond.get_atom1_id(),
                atom2_id=bond.get_atom2_id(),
            )
            assert np.isclose(test, bond_length, atol=1e-6)



import numpy as np
from mchammer import get_atom_distance


def test_get_atom_distance() -> None:
    position_matrix = np.array(
        [
            [0, 0, 0],
            [0, 1, 0],
            [2, 0, 0],
        ]
    )
    assert get_atom_distance(position_matrix, 0, 1) == 1
    assert get_atom_distance(position_matrix, 0, 2) == 2



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import numpy as np
import pytest

if TYPE_CHECKING:
    import mchammer as mch


def test_molecule_get_position_matrix(
    molecule: mch.Molecule,
    position_matrix: np.ndarray,
) -> None:
    assert np.all(
        np.equal(
            position_matrix,
            molecule.get_position_matrix(),
        )
    )


def test_molecule_with_position_matrix(
    molecule: mch.Molecule,
    position_matrix2: np.ndarray,
) -> None:
    test = molecule.with_position_matrix(position_matrix2)
    assert np.all(
        np.equal(
            position_matrix2,
            test.get_position_matrix(),
        )
    )


def test_molecule_with_displacement(
    molecule: mch.Molecule,
    displacement: np.ndarray,
    displaced_position_matrix: np.ndarray,
) -> None:
    test = molecule.with_displacement(displacement)
    print(test.get_position_matrix())
    assert np.all(
        np.allclose(
            displaced_position_matrix,
            test.get_position_matrix(),
        )
    )


@pytest.fixture()
def path(tmpdir) -> None:  # noqa: ANN001
    return os.path.join(tmpdir, "molecule.xyz")  # noqa: PTH118


def test_molecule_write_xyz_file(molecule: mch.Molecule, path: str) -> None:
    molecule.write_xyz_file(path)
    content = molecule.write_xyz_content()
    with open(path) as f:
        test_lines = f.readlines()

    assert "".join(test_lines) == "".join(content)


def test_molecule_get_atoms(
    molecule: mch.Molecule,
    atoms: list[mch.Atom],
) -> None:
    for test, atom in zip(molecule.get_atoms(), atoms):
        assert test.get_id() == atom.get_id()
        assert test.get_element_string() == atom.get_element_string()


def test_molecule_get_bonds(
    molecule: mch.Molecule,
    bonds: list[mch.Bond],
) -> None:
    for test, bond in zip(molecule.get_bonds(), bonds):
        assert test.get_id() == bond.get_id()
        assert test.get_atom1_id() == bond.get_atom1_id()
        assert test.get_atom2_id() == bond.get_atom2_id()


def test_molecule_get_centroid(
    molecule: mch.Molecule,
    position_matrix: np.ndarray,
    centroid: np.ndarray,
) -> None:
    test = molecule.with_position_matrix(position_matrix)
    assert np.all(
        np.allclose(
            centroid,
            test.get_centroid(),
            atol=1e-6,
        )
    )


def test_molecule_get_subunits(molecule: mch.Molecule, subunits: dict) -> None:
    test = molecule.get_subunits(bond_pair_ids=((0, 3),))
    assert test == subunits


def test_molecule_get_num_atoms(
    molecule: mch.Molecule,
    num_atoms: int,
) -> None:
    assert molecule.get_num_atoms() == num_atoms



def test_bond_get_id(bond_info: tuple) -> None:
    assert bond_info[0].get_id() == bond_info[1]
    assert bond_info[0].id == bond_info[1]


def test_bond_get_atom1_id(bond_info: tuple) -> None:
    assert bond_info[0].get_atom1_id() == bond_info[2]
    assert bond_info[0].atom1_id == bond_info[2]


def test_bond_get_atom2_id(bond_info: tuple) -> None:
    assert bond_info[0].get_atom2_id() == bond_info[3]
    assert bond_info[0].atom2_id == bond_info[3]



"""Module of Monte Carlo operations."""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING

import numpy as np
from scipy.spatial.transform import Rotation

if TYPE_CHECKING:
    from .molecule import Molecule


def translate_atoms_along_vector(
    mol: Molecule,
    atom_ids: tuple[int, ...],
    vector: np.ndarray,
) -> Molecule:
    """Translate atoms with given ids along a vector."""
    new_position_matrix = deepcopy(mol.get_position_matrix())
    for atom in mol.get_atoms():
        if atom.get_id() not in atom_ids:
            continue
        pos = mol.get_position_matrix()[atom.get_id()]
        new_position_matrix[atom.get_id()] = pos - vector

    return mol.with_position_matrix(new_position_matrix)


def translate_molecule_along_vector(
    mol: Molecule,
    vector: np.ndarray,
) -> Molecule:
    """Translate a whole molecule along a vector."""
    return mol.with_displacement(vector)


def test_move(
    beta: float,
    curr_pot: float,
    new_pot: float,
    generator: np.random.Generator,
) -> bool:
    """Test an MC move based on boltzmann energy."""
    if new_pot < curr_pot:
        return True

    exp_term = np.exp(-beta * (new_pot - curr_pot))
    rand_number = generator.random()

    return exp_term > rand_number


def rotation_matrix_arbitrary_axis(
    angle: float,
    axis: np.ndarray,
) -> np.ndarray:
    """Returns a rotation matrix of `angle` radians about `axis`.

    Parameters:
        angle:
            The size of the rotation in radians.

        axis:
            A 3 element aray which represents a vector. The vector is the
            axis about which the rotation is carried out. Must be of
            unit magnitude.

    Returns:
        A ``3x3`` array representing a rotation matrix.
    """
    a = np.cos(angle / 2)
    b, c, d = axis * np.sin(angle / 2)

    e11 = np.square(a) + np.square(b) - np.square(c) - np.square(d)
    e12 = 2 * (b * c - a * d)
    e13 = 2 * (b * d + a * c)

    e21 = 2 * (b * c + a * d)
    e22 = np.square(a) + np.square(c) - np.square(b) - np.square(d)
    e23 = 2 * (c * d - a * b)

    e31 = 2 * (b * d - a * c)
    e32 = 2 * (c * d + a * b)
    e33 = np.square(a) + np.square(d) - np.square(b) - np.square(c)

    # Initialize as a scipy Rotation object, which normalizes the
    # matrix and allows for returns as quaternion or alternative
    # type in the future.
    return Rotation.from_matrix(
        np.array([[e11, e12, e13], [e21, e22, e23], [e31, e32, e33]])
    ).as_matrix()


def rotate_molecule_by_angle(
    mol: Molecule,
    angle: float,
    axis: np.ndarray,
    origin: np.ndarray,
) -> Molecule:
    """Rotate a molecule by an angle about an axis and origin."""
    new_position_matrix = mol.get_position_matrix()
    # Set the origin of the rotation to "origin".
    new_position_matrix = new_position_matrix - origin
    # Perform rotation.
    rot_mat = rotation_matrix_arbitrary_axis(angle, axis)
    # Apply the rotation matrix on the position matrix, to get the
    # new position matrix.
    new_position_matrix = (rot_mat @ new_position_matrix.T).T
    # Return the centroid of the molecule to the original position.
    new_position_matrix = new_position_matrix + origin

    return mol.with_position_matrix(new_position_matrix)



"""MCHammer package."""

from mchammer._internal.atom import Atom
from mchammer._internal.bond import Bond
from mchammer._internal.collapser import Collapser
from mchammer._internal.mc_operations import (
    rotate_molecule_by_angle,
    rotation_matrix_arbitrary_axis,
    test_move,
    translate_atoms_along_vector,
    translate_molecule_along_vector,
)
from mchammer._internal.molecule import Molecule
from mchammer._internal.optimizer import Optimizer
from mchammer._internal.radii import get_radius
from mchammer._internal.results import MCStepResult, Result, StepResult
from mchammer._internal.utilities import get_atom_distance, get_bond_vector

__all__ = [
    "Atom",
    "Bond",
    "Molecule",
    "get_radius",
    "StepResult",
    "MCStepResult",
    "Result",
    "get_atom_distance",
    "Optimizer",
    "Collapser",
    "get_bond_vector",
    "translate_atoms_along_vector",
    "translate_molecule_along_vector",
    "test_move",
    "rotate_molecule_by_angle",
    "rotation_matrix_arbitrary_axis",
]



import stk

bb1 = stk.BuildingBlock("NCCN", [stk.PrimaryAminoFactory()])
bb2 = stk.BuildingBlock("O=CCC=O", [stk.AldehydeFactory()])

polymer = stk.ConstructedMolecule(
    topology_graph=stk.polymer.Linear(
        building_blocks=(bb1, bb2),
        repeating_unit="AB",
        num_repeating_units=6,
        optimizer=stk.MCHammer(),
    ),
)
polymer.write("mchammer.mol")


bb1 = stk.BuildingBlock("NCCN", [stk.PrimaryAminoFactory()])
bb2 = stk.BuildingBlock("O=CCC=O", [stk.AldehydeFactory()])

polymer = stk.ConstructedMolecule(
    topology_graph=stk.polymer.Linear(
        building_blocks=(bb1, bb2),
        repeating_unit="AB",
        num_repeating_units=2,
        optimizer=stk.Collapser(),
    ),
)
polymer.write("collapser.mol")



"""Define classes holding result information."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections import abc

    import numpy as np


@dataclass
class StepResult:
    """Results of a step.

    Parameters:
        step:
            Step number.

        position_matrix:
            A position matrix after performing this step. The shape of
            the matrix is ``(n, 3)``.

        max_bond_distance:
            Max length of bonds to be optimized in Angstrom.

        log:
            String log of this step.

    """

    step: int
    position_matrix: np.ndarray
    max_bond_distance: float
    log: str

    def get_properties(self) -> dict:
        """Get step properties."""
        return {"max_bond_distance": self.max_bond_distance}


@dataclass
class MCStepResult(StepResult):
    """Results of a step.

    Parameters:
        step:
            Step number.

        position_matrix:
            A position matrix after performing this step. The shape of
            the matrix is ``(n, 3)``.

        passed:
            Flag for whether the MC move passed, or was reverted to
            the previous step.

        system_potential:
            System potential of the structure after this step.

        nonbonded_potential:
            Nonbonded potential of the structure after this step.

        max_bond_distance:
            Max length of bonds to be optimized in Angstrom.

        log:
            String log of this step.

    """

    step: int
    log: str
    position_matrix: np.ndarray
    system_potential: float
    nonbonded_potential: float
    max_bond_distance: float
    passed: bool | None = None

    def get_properties(self) -> dict:
        """Get step properties."""
        return {
            "max_bond_distance": self.max_bond_distance,
            "system_potential": self.system_potential,
            "nonbonded_potential": self.nonbonded_potential,
            "passed": self.passed,
        }


class Result:
    """Result of optimization."""

    def __init__(self, start_time: float) -> None:
        """Initialize a :class:`Result` instance.

        Parameters:
            start_time:
                Start of run timing.

        """
        self._start_time = start_time
        self._log = ""
        self._step_results: dict[int, StepResult] = {}
        self._step_count = 0

    def add_step_result(self, step_result: StepResult) -> None:
        """Add StepResult."""
        self._step_results[step_result.step] = step_result
        self.update_log(step_result.log)
        self._step_count = step_result.step

    def update_log(self, string: str) -> None:
        """Update result log."""
        self._log += string

    def get_log(self) -> str:
        """Get result log."""
        return self._log

    def get_number_passed(self) -> int:
        """Get Number of steps that passed."""
        return len(
            [
                1
                for step in self._step_results
                if self._step_results[step].passed  # type: ignore[attr-defined]
            ]
        )

    def get_final_position_matrix(self) -> np.ndarray:
        """Get final molecule in result."""
        return self._step_results[self._step_count].position_matrix

    def get_timing(self, time: float) -> float:
        """Get run timing."""
        return time - self._start_time

    def get_step_count(self) -> int:
        """Get step count."""
        return self._step_count

    def get_steps_properties(self) -> abc.Iterable[tuple[int, dict]]:
        """Yield properties of all steps."""
        for step in self._step_results:
            yield step, self._step_results[step].get_properties()

    def get_trajectory(self) -> abc.Iterable[tuple[int, np.ndarray]]:
        """Yield .Molecule of all steps."""
        for step in self._step_results:
            yield step, self._step_results[step].position_matrix



import pathlib

import mchammer as mch
import numpy as np
import stk

from .utilities import get_long_bond_ids, get_mch_bonds, get_subunits


def test_stk_collapser() -> None:
    bb1 = stk.BuildingBlock("NCCN", [stk.PrimaryAminoFactory()])
    bb2 = stk.BuildingBlock("O=CCC=O", [stk.AldehydeFactory()])

    polymer_opt = stk.ConstructedMolecule(
        topology_graph=stk.polymer.Linear(
            building_blocks=(bb1, bb2),
            repeating_unit="AB",
            num_repeating_units=2,
            optimizer=stk.Collapser(),
        ),
    )

    polymer_unopt = stk.ConstructedMolecule(
        topology_graph=stk.polymer.Linear(
            building_blocks=(bb1, bb2),
            repeating_unit="AB",
            num_repeating_units=2,
        ),
    )
    mch_mol = mch.Molecule(
        atoms=(
            mch.Atom(
                id=atom.get_id(),
                element_string=atom.__class__.__name__,
            )
            for atom in polymer_unopt.get_atoms()
        ),
        bonds=get_mch_bonds(polymer_unopt),
        position_matrix=polymer_unopt.get_position_matrix(),
    )

    # Run optimization.
    optimizer = mch.Collapser(
        step_size=0.1,
        distance_threshold=1.5,
        scale_steps=True,
    )
    mch_mol, result = optimizer.get_result(
        mol=mch_mol,
        bond_pair_ids=tuple(get_long_bond_ids(polymer_unopt)),
        subunits=get_subunits(polymer_unopt),
    )
    new_polymer = polymer_unopt.with_position_matrix(
        position_matrix=mch_mol.get_position_matrix()
    )

    assert np.all(
        np.equal(
            polymer_opt.get_position_matrix(),
            new_polymer.get_position_matrix(),
        )
    )

    known = stk.BuildingBlock.init_from_file(
        pathlib.Path(__file__).resolve().parent / "collapser.mol"
    )

    assert np.all(
        np.equal(
            known.get_position_matrix(),
            new_polymer.get_position_matrix(),
        )
    )


def test_stk_mchammer() -> None:
    bb1 = stk.BuildingBlock("NCCN", [stk.PrimaryAminoFactory()])
    bb2 = stk.BuildingBlock("O=CCC=O", [stk.AldehydeFactory()])

    polymer_opt = stk.ConstructedMolecule(
        topology_graph=stk.polymer.Linear(
            building_blocks=(bb1, bb2),
            repeating_unit="AB",
            num_repeating_units=6,
            optimizer=stk.MCHammer(),
        ),
    )

    polymer_unopt = stk.ConstructedMolecule(
        topology_graph=stk.polymer.Linear(
            building_blocks=(bb1, bb2),
            repeating_unit="AB",
            num_repeating_units=6,
        ),
    )
    mch_mol = mch.Molecule(
        atoms=(
            mch.Atom(
                id=atom.get_id(),
                element_string=atom.__class__.__name__,
            )
            for atom in polymer_unopt.get_atoms()
        ),
        bonds=tuple(get_mch_bonds(polymer_unopt)),
        position_matrix=polymer_unopt.get_position_matrix(),
    )

    # Run optimization.
    optimizer = mch.Optimizer(
        step_size=0.25,
        target_bond_length=1.2,
        num_steps=500,
        bond_epsilon=50,
        nonbond_epsilon=20,
        nonbond_sigma=1.2,
        nonbond_mu=3,
        beta=2,
        random_seed=1000,
    )
    mch_mol, _ = optimizer.get_result(
        mol=mch_mol,
        bond_pair_ids=tuple(get_long_bond_ids(polymer_unopt)),
        subunits=get_subunits(polymer_unopt),
    )
    new_polymer = polymer_unopt.with_position_matrix(
        position_matrix=mch_mol.get_position_matrix()
    )

    assert np.all(
        np.equal(
            polymer_opt.get_position_matrix(),
            new_polymer.get_position_matrix(),
        )
    )

    known = stk.BuildingBlock.init_from_file(
        pathlib.Path(__file__).resolve().parent / "mchammer.mol"
    )

    assert np.all(
        np.equal(
            known.get_position_matrix(),
            new_polymer.get_position_matrix(),
        )
    )



def test_atom_get_id(atom_info: tuple) -> None:
    assert atom_info[0].get_id() == atom_info[1]
    assert atom_info[0].id == atom_info[1]


def test_atom_get_element_string(atom_info: tuple) -> None:
    assert atom_info[0].get_element_string() == atom_info[2]
    assert atom_info[0].element_string == atom_info[2]


def test_atom_get_radius(atom_info: tuple) -> None:
    assert atom_info[0].get_radius() == atom_info[3]
    assert atom_info[0].radius == atom_info[3]



"""Optimizer for minimising intermolecular distances."""

from __future__ import annotations

import time
from itertools import combinations, product
from typing import TYPE_CHECKING

import numpy as np

from .results import Result, StepResult
from .utilities import get_atom_distance

if TYPE_CHECKING:
    from collections import abc

    from .molecule import Molecule


class Collapser:
    """Moves rigid-subunits toward center of mass of molecule."""

    def __init__(
        self,
        step_size: float,
        distance_threshold: float,
        scale_steps: bool = True,  # noqa: FBT001, FBT002
    ) -> None:
        """Initialize a :class:`Collapser` instance.

        Parameters:
            step_size:
                The relative size of the step to take during collapse.

            distance_threshold:
                Distance between distinct subunits to use as
                threshold for halting collapse in Angstrom.

            scale_steps:
                Whether to scale the step of each distict building block
                by their relative distance from the molecules centroid.
                Defaults to ``True``

        """
        self._step_size = step_size
        self._distance_threshold = distance_threshold
        self._scale_steps = scale_steps

    def _has_short_contacts(self, mol: Molecule, subunits: dict) -> bool:
        """Calculate if there are short contants in mol."""
        return any(
            dist < self._distance_threshold
            for dist in self._get_subunit_distances(mol, subunits)
        )

    def _get_subunit_distances(
        self,
        mol: Molecule,
        subunits: dict[int, tuple],
    ) -> abc.Iterable[float]:
        """Yield the distances between subunits in mol.

        Ignores H atoms.

        """
        atom_elements = [i.get_element_string() for i in mol.get_atoms()]
        position_matrix = mol.get_position_matrix()

        for su1, su2 in combinations(subunits, 2):
            if su1 == su2:
                continue
            su1_atom_ids = subunits[su1]
            su2_atom_ids = subunits[su2]
            for atom1_id, atom2_id in product(su1_atom_ids, su2_atom_ids):
                atom1_element = atom_elements[atom1_id]
                atom2_element = atom_elements[atom2_id]
                if atom1_element != "H" and atom2_element != "H":
                    yield get_atom_distance(
                        position_matrix=position_matrix,
                        atom1_id=atom1_id,
                        atom2_id=atom2_id,
                    )

    def _get_new_position_matrix(  # noqa: PLR0913
        self,
        mol: Molecule,
        subunits: dict,
        vectors: dict,
        scales: dict,
        step_size: float | None,
    ) -> np.ndarray:
        """Get the position matrix of the mol after translation."""
        if step_size is None:
            step_size = self._step_size

        new_position_matrix = mol.get_position_matrix()
        for su in subunits:
            for atom_id in subunits[su]:
                pos = mol.get_position_matrix()[atom_id]
                new_position_matrix[atom_id] = (
                    pos - step_size * vectors[su] * scales[su]
                )

        return new_position_matrix

    def _get_subunit_vectors(
        self,
        mol: Molecule,
        subunits: dict,
    ) -> tuple[dict, dict]:
        """Get the subunit to COM vectors.

        Parameters:
            mol:
                The molecule to be optimized.

            subunits:
                The subunits of `mol` split by bonds defined by
                `bond_pair_ids`. Key is subunit identifier, Value is
                :class:`iterable` of atom ids in subunit.

        Returns:
            su_cent_vectors:
                :class:`dict` mapping :class:`int`: to
                :class:`numpy.ndarray`
                Dictionary mapping subunit ids (keys) to centroid
                vectors (values) of each distinct building block in the
                molecule.

            su_cent_scales:
                :class:`dict` mapping :class:`int`: to :class:`float`
                Dictionary mapping subunit ids (keys) to relative
                magnitude of centroid vectors (values) of each distinct
                building block in the molecule.

        """
        centroid = mol.get_centroid()

        # Get subunit centroid to molecule centroid vector.

        su_cent_vectors = {
            i: mol.get_centroid(atom_ids=subunits[i]) - centroid
            for i in subunits
        }

        # Scale the step size based on the different distances of
        # subunits from the COM. Impacts anisotropic topologies.
        if self._scale_steps:
            norms = {
                i: np.linalg.norm(su_cent_vectors[i]) for i in su_cent_vectors
            }
            max_distance = max(list(norms.values()))
            su_cent_scales = {i: float(norms[i] / max_distance) for i in norms}
        else:
            su_cent_scales = {i: 1.0 for i in su_cent_vectors}

        return su_cent_vectors, su_cent_scales

    def _run_step(  # noqa: PLR0913
        self,
        mol: Molecule,
        bond_pair_ids: tuple,
        subunits: dict,
        step: int,
        step_size: float | None = None,
    ) -> tuple[Molecule, StepResult]:
        # Get the subunit to centroid vectors and their relative scale.
        su_cent_vectors, su_cent_scales = self._get_subunit_vectors(
            mol=mol, subunits=subunits
        )

        new_position_matrix = self._get_new_position_matrix(
            mol=mol,
            subunits=subunits,
            vectors=su_cent_vectors,
            scales=su_cent_scales,
            step_size=step_size,
        )

        mol = mol.with_position_matrix(new_position_matrix)
        # Update properties at each step.
        max_bond_distance = max(
            [
                get_atom_distance(
                    position_matrix=mol.get_position_matrix(),
                    atom1_id=bond[0],
                    atom2_id=bond[1],
                )
                for bond in bond_pair_ids
            ]
        )
        step_result = StepResult(
            step=step,
            position_matrix=new_position_matrix,
            max_bond_distance=max_bond_distance,
            log=f"{step} {max_bond_distance}\n",
        )
        return mol, step_result

    def get_trajectory(
        self,
        mol: Molecule,
        bond_pair_ids: tuple,
        subunits: dict,
    ) -> tuple[Molecule, Result]:
        """Get trajectory of optimization run on `mol`.

        Parameters:
            mol:
                The molecule to be optimized.

            bond_pair_ids:
                :class:`iterable` of :class:`tuple` of :class:`ints`
                Iterable of pairs of atom ids with bond between them to
                optimize.

            subunits:
                The subunits of `mol` split by bonds defined by
                `bond_pair_ids`. Key is subunit identifier, Value is
                :class:`iterable` of atom ids in subunit.

        Returns:
            mol:
                The optimized molecule.

            result:
                The result of the optimization including all steps.

        """
        result = Result(start_time=time.time())
        result.update_log(
            "====================================================\n"
            f"                Collapser algorithm.\n"
            "====================================================\n"
            f"Step size: {self._step_size}\n"
            f"Scale steps?: {self._scale_steps}\n"
            f"Distance threshold: {self._distance_threshold}\n"
        )
        result.update_log(
            f"There are {len(bond_pair_ids)} bonds to optimize.\n"
        )
        result.update_log(
            f"There are {len(subunits)} sub units with N atoms:\n"
            f"{[len(subunits[i]) for i in subunits]}\n"
        )
        result.update_log(
            "====================================================\n"
            "                 Running optimisation!              \n"
            "====================================================\n\n"
        )

        step = 0
        while not self._has_short_contacts(mol, subunits):
            step += 1
            mol, step_result = self._run_step(
                mol=mol,
                bond_pair_ids=bond_pair_ids,
                subunits=subunits,
                step=step,
            )
            result.add_step_result(step_result=step_result)

        # Check that we have not gone too far.
        min_dist = min(
            dist for dist in self._get_subunit_distances(mol, subunits)
        )
        if min_dist < self._distance_threshold / 2:
            step += 1
            mol, step_result = self._run_step(
                mol=mol,
                bond_pair_ids=bond_pair_ids,
                subunits=subunits,
                step=step,
                step_size=-(self._step_size / 2),
            )
            result.add_step_result(step_result=step_result)

        result.update_log(
            "====================================================\n"
            f"Steps run: {step}\n"
            f"Minimum inter-subunit distance: {min_dist}\n"
            "====================================================\n"
        )

        return mol, result

    def get_result(
        self,
        mol: Molecule,
        bond_pair_ids: tuple,
        subunits: dict,
    ) -> tuple[Molecule, StepResult]:
        """Get final result of the optimization of `mol`.

        Parameters:
            mol:
                The molecule to be optimized.

            bond_pair_ids:
                :class:`iterable` of :class:`tuple` of :class:`ints`
                Iterable of pairs of atom ids with bond between them to
                optimize.

            subunits:
                The subunits of `mol` split by bonds defined by
                `bond_pair_ids`. Key is subunit identifier, Value is
                :class:`iterable` of atom ids in subunit.

        Returns:
            mol:
                The optimized molecule.

            result:
                The result of the final optimization step.

        """
        step = 0
        while not self._has_short_contacts(mol, subunits):
            step += 1
            mol, step_result = self._run_step(
                mol=mol,
                bond_pair_ids=bond_pair_ids,
                subunits=subunits,
                step=step,
            )

        # Check that we have not gone too far.
        min_dist = min(
            dist for dist in self._get_subunit_distances(mol, subunits)
        )
        if min_dist < self._distance_threshold / 2:
            step += 1
            mol, step_result = self._run_step(
                mol=mol,
                bond_pair_ids=bond_pair_ids,
                subunits=subunits,
                step=step,
                step_size=-(self._step_size / 2),
            )
        return mol, step_result



"""Bond class."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Bond:
    """Bond between two atoms.

    Parameters:
        id:
            ID to be assigned to bond.

        atom_ids:
            IDs of atom 1 and atom 2 in bond, where atom 1 is always
            the smaller number and the IDs cannot be the same.

    """

    id: int
    atom_ids: tuple[int, int]

    def __post_init__(self) -> None:
        """Post initialization of bond."""
        if len(set(self.atom_ids)) != 2:  # noqa: PLR2004
            msg = "Two distinct atom ids are required."
            raise ValueError(msg)
        self.atom1_id, self.atom2_id = sorted(self.atom_ids)

    def get_id(self) -> int:
        """Get bond ID."""
        return self.id

    def get_atom1_id(self) -> int:
        """Get ID of atom 1 in bond."""
        return self.atom1_id

    def get_atom2_id(self) -> int:
        """Get ID of atom 2 in bond."""
        return self.atom2_id

    def __str__(self) -> str:
        """String representation of Bond."""
        return repr(self)

    def __repr__(self) -> str:
        """String representation of Bond."""
        return (
            f"{self.__class__.__name__}(id={self.get_id()}, "
            f"atom1_id={self.get_atom1_id()}, "
            f"atom2_id={self.get_atom2_id()})"
        )



"""This module defines general-purpose objects, functions and classes."""

from __future__ import annotations

from typing import TYPE_CHECKING

from scipy.spatial.distance import euclidean

if TYPE_CHECKING:
    import numpy as np


def get_atom_distance(
    position_matrix: np.ndarray,
    atom1_id: int,
    atom2_id: int,
) -> float:
    """Return the distance between two atoms."""
    return float(
        euclidean(u=position_matrix[atom1_id], v=position_matrix[atom2_id])
    )


def get_bond_vector(
    position_matrix: np.ndarray,
    bond_pair: tuple[int, int],
) -> np.ndarray:
    """Get vector from atom1 to atom2 in bond."""
    atom1_pos = position_matrix[bond_pair[0]]
    atom2_pos = position_matrix[bond_pair[1]]
    return atom2_pos - atom1_pos



"""Optimizer for minimising intermolecular distances."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING

import numpy as np
from scipy.spatial.distance import pdist

from .mc_operations import test_move, translate_atoms_along_vector
from .results import MCStepResult, Result
from .utilities import get_atom_distance, get_bond_vector

if TYPE_CHECKING:
    from .molecule import Molecule


class Optimizer:
    """Optimize target bonds using MC algorithm.

    A Metropolis MC algorithm is applied to perform rigid
    translations of the subunits separatred by the target bonds.

    """

    def __init__(  # noqa: PLR0913
        self,
        step_size: float,
        target_bond_length: float,
        num_steps: int,
        bond_epsilon: float = 50,
        nonbond_epsilon: float = 20,
        nonbond_sigma: float = 1.2,
        nonbond_mu: float = 3,
        beta: float = 2,
        random_seed: int | None = 1000,
    ) -> None:
        """Initialize a :class:`Optimizer` instance.

        Parameters:
            step_size:
                The relative size of the step to take during step.

            target_bond_length:
                Target equilibrium bond length for long bonds to minimize
                to.

            num_steps:
                Number of MC moves to perform.

            bond_epsilon:
                Value of epsilon used in the bond potential in MC moves.
                Determines strength of the bond potential.
                Defaults to 50.

            nonbond_epsilon:
                Value of epsilon used in the nonbond potential in MC moves.
                Determines strength of the nonbond potential.
                Defaults to 20.

            nonbond_sigma:
                Value of sigma used in the nonbond potential in MC moves.
                Defaults to 1.2.

            nonbond_mu:
                Value of mu used in the nonbond potential in MC moves.
                Determines the steepness of the nonbond potential.
                Defaults to 3.

            beta:
                Value of beta used in the in MC moves. Beta takes the
                place of the inverse boltzmann temperature.
                Defaults to 2.

            random_seed:
                Random seed to use for MC algorithm. Should only be set to
                ``None`` if system-based random seed is desired. Defaults
                to a set seed of 1000, to avoid randomness.

        """
        self._step_size = step_size
        self._target_bond_length = target_bond_length
        self._num_steps = num_steps
        self._bond_epsilon = bond_epsilon
        self._nonbond_epsilon = nonbond_epsilon
        self._nonbond_sigma = nonbond_sigma
        self._nonbond_mu = nonbond_mu
        self._beta = beta
        if random_seed is None:
            self._generator = np.random.default_rng()
        else:
            self._generator = np.random.default_rng(random_seed)

    def _bond_potential(self, distance: float) -> float:
        """Define an arbitrary parabolic bond potential.

        This potential has no relation to an empircal forcefield.

        """
        potential = (distance - self._target_bond_length) ** 2
        return self._bond_epsilon * potential

    def _nonbond_potential(self, distance: float) -> float:
        """Define an arbitrary repulsive nonbonded potential.

        This potential has no relation to an empircal forcefield.

        """
        return self._nonbond_epsilon * (
            (self._nonbond_sigma / distance) ** self._nonbond_mu
        )

    def _compute_nonbonded_potential(
        self,
        position_matrix: np.ndarray,
    ) -> float:
        # Get all pairwise distances between atoms in each subunut.
        pair_dists = pdist(position_matrix)
        return np.sum(self._nonbond_potential(pair_dists))

    def compute_potential(
        self,
        mol: Molecule,
        bond_pair_ids: list,
    ) -> tuple[float, float]:
        """Compute the potential of a molecule."""
        position_matrix = mol.get_position_matrix()
        nonbonded_potential = self._compute_nonbonded_potential(
            position_matrix=position_matrix,
        )
        system_potential = nonbonded_potential
        for bond in bond_pair_ids:
            system_potential += self._bond_potential(
                distance=get_atom_distance(
                    position_matrix=position_matrix,
                    atom1_id=bond[0],
                    atom2_id=bond[1],
                )
            )

        return system_potential, nonbonded_potential

    def _output_top_lines(self) -> str:
        return (
            "====================================================\n"
            "                MCHammer optimisation               \n"
            "                ---------------------               \n"
            "                    Andrew Tarzia                   \n"
            "                ---------------------               \n"
            "                                                    \n"
            "Settings:                                           \n"
            f" step size = {self._step_size} \n"
            f" target bond length = {self._target_bond_length} \n"
            f" num. steps = {self._num_steps} \n"
            f" bond epsilon = {self._bond_epsilon} \n"
            f" nonbond epsilon = {self._nonbond_epsilon} \n"
            f" nonbond sigma = {self._nonbond_sigma} \n"
            f" nonbond mu = {self._nonbond_mu} \n"
            f" beta = {self._beta} \n"
            "====================================================\n\n"
        )

    def _run_first_step(
        self,
        mol: Molecule,
        bond_pair_ids: list,
    ) -> tuple[Molecule, MCStepResult]:
        system_potential, nonbonded_potential = self.compute_potential(
            mol=mol, bond_pair_ids=bond_pair_ids
        )

        # Update properties at each step.
        max_bond_distance = max(
            [
                get_atom_distance(
                    position_matrix=mol.get_position_matrix(),
                    atom1_id=bond[0],
                    atom2_id=bond[1],
                )
                for bond in bond_pair_ids
            ]
        )
        step_result = MCStepResult(
            step=0,
            position_matrix=mol.get_position_matrix(),
            passed=None,
            system_potential=system_potential,
            nonbonded_potential=nonbonded_potential,
            max_bond_distance=max_bond_distance,
            log=(
                f"{0} "
                f"{system_potential} "
                f"{nonbonded_potential} "
                f"{max_bond_distance} "
                "-- --\n"
            ),
        )

        return mol, step_result

    def _run_step(  # noqa: PLR0913
        self,
        mol: Molecule,
        bond_pair_ids: list,
        subunits: dict,
        step: int,
        system_potential: float,
        nonbonded_potential: float,
    ) -> tuple[Molecule, MCStepResult]:
        position_matrix = mol.get_position_matrix()

        # Randomly select a bond to optimize from bonds.
        bond_ids = self._generator.choice(bond_pair_ids)
        bond_vector = get_bond_vector(
            position_matrix=position_matrix, bond_pair=bond_ids
        )

        # Get subunits connected by selected bonds.
        subunit_1 = next(i for i in subunits if bond_ids[0] in subunits[i])
        subunit_2 = next(i for i in subunits if bond_ids[1] in subunits[i])

        # Choose subunit to move out of the two connected by the
        # bond randomly.
        moving_su = self._generator.choice([subunit_1, subunit_2])
        moving_su_atom_ids = tuple(i for i in subunits[moving_su])

        # Random number from -1 to 1 for multiplying translation.
        rand = (self._generator.random() - 0.5) * 2
        # Define translation along long bond vector where
        # direction is from force, magnitude is randomly
        # scaled.
        bond_translation = -bond_vector * self._step_size * rand

        # Define subunit COM vector to molecule COM.
        cent = mol.get_centroid()
        su_cent_vector = mol.get_centroid(atom_ids=moving_su_atom_ids) - cent
        com_translation = su_cent_vector * self._step_size * rand

        # Randomly choose between translation along long bond
        # vector or along BB-COM vector.
        translation_vector = self._generator.choice(
            [bond_translation, com_translation]  # type: ignore[arg-type]
        )

        # Translate building block.
        # Update atom position of building block.
        mol = translate_atoms_along_vector(
            mol=mol,
            atom_ids=moving_su_atom_ids,
            vector=translation_vector,  # type: ignore[arg-type]
        )

        (
            new_system_potential,
            new_nonbonded_potential,
        ) = self.compute_potential(mol=mol, bond_pair_ids=bond_pair_ids)

        if test_move(
            beta=self._beta,
            curr_pot=system_potential,
            new_pot=new_system_potential,
            generator=self._generator,
        ):
            updated = "T"
            passed = True
            system_potential = new_system_potential
            nonbonded_potential = new_nonbonded_potential
        else:
            updated = "F"
            passed = False
            # Reverse move.
            mol = translate_atoms_along_vector(
                mol=mol,
                atom_ids=moving_su_atom_ids,
                vector=-translation_vector,  # type: ignore[operator]
            )

        # Update properties at each step.
        max_bond_distance = max(
            [
                get_atom_distance(
                    position_matrix=mol.get_position_matrix(),
                    atom1_id=bond[0],
                    atom2_id=bond[1],
                )
                for bond in bond_pair_ids
            ]
        )
        step_result = MCStepResult(
            step=step,
            position_matrix=mol.get_position_matrix(),
            passed=passed,
            system_potential=system_potential,
            nonbonded_potential=nonbonded_potential,
            max_bond_distance=max_bond_distance,
            log=(
                f"{step} "
                f"{system_potential} "
                f"{nonbonded_potential} "
                f"{max_bond_distance} "
                f"{bond_ids} {updated}\n"
            ),
        )

        return mol, step_result

    def get_trajectory(
        self,
        mol: Molecule,
        bond_pair_ids: list[tuple],
        subunits: dict,
    ) -> tuple[Molecule, Result]:
        """Get trajectory of optimization run on `mol`.

        Parameters:
            mol:
                The molecule to be optimized.

            bond_pair_ids:
                :class:`iterable` of :class:`tuple` of :class:`ints`
                Iterable of pairs of atom ids with bond between them to
                optimize.

            subunits:
                The subunits of `mol` split by bonds defined by
                `bond_pair_ids`. Key is subunit identifier, Value is
                :class:`iterable` of atom ids in subunit.

        Returns:
            mol:
                The optimized molecule.

            result:
                The result of the optimization including all steps.

        """
        result = Result(start_time=time.time())

        result.update_log(self._output_top_lines())
        result.update_log(
            f"There are {len(bond_pair_ids)} bonds to optimize.\n"
        )
        result.update_log(
            f"There are {len(subunits)} sub units with N atoms:\n"
            f"{[len(subunits[i]) for i in subunits]}\n"
        )
        result.update_log(
            "====================================================\n"
            "                 Running optimisation!              \n"
            "====================================================\n\n"
        )
        result.update_log(
            "step system_potential nonbond_potential max_dist "
            "opt_bbs updated?\n"
        )
        mol, step_result = self._run_first_step(mol, bond_pair_ids)
        system_potential = step_result.system_potential
        nonbonded_potential = step_result.nonbonded_potential
        result.add_step_result(step_result=step_result)

        for step in range(1, self._num_steps):
            mol, step_result = self._run_step(
                mol=mol,
                bond_pair_ids=bond_pair_ids,
                subunits=subunits,
                step=step,
                system_potential=system_potential,
                nonbonded_potential=nonbonded_potential,
            )

            system_potential = step_result.system_potential
            nonbonded_potential = step_result.nonbonded_potential
            result.add_step_result(step_result=step_result)

        num_passed = result.get_number_passed()
        result.update_log(
            string=(
                "\n============================================\n"
                "Optimisation done:\n"
                f"{num_passed} steps passed: "
                f"{(num_passed/self._num_steps)*100}"
                "%\n"
                f"Total optimisation time: "
                f"{round(result.get_timing(time.time()), 4)}s\n"
                "============================================\n"
            ),
        )

        return mol, result

    def get_result(
        self,
        mol: Molecule,
        bond_pair_ids: list,
        subunits: dict,
    ) -> tuple[Molecule, MCStepResult]:
        """Get final result of optimization run on `mol`.

        Parameters:
            mol:
                The molecule to be optimized.

            bond_pair_ids:
                :class:`iterable` of :class:`tuple` of :class:`ints`
                Iterable of pairs of atom ids with bond between them to
                optimize.

            subunits:
                The subunits of `mol` split by bonds defined by
                `bond_pair_ids`. Key is subunit identifier, Value is
                :class:`iterable` of atom ids in subunit.

        Returns:
            mol:
                The optimized molecule.

            result:
                The result of the final optimization step.

        """
        mol, step_result = self._run_first_step(mol, bond_pair_ids)
        system_potential = step_result.system_potential
        nonbonded_potential = step_result.nonbonded_potential

        for step in range(1, self._num_steps):
            mol, step_result = self._run_step(
                mol=mol,
                bond_pair_ids=bond_pair_ids,
                subunits=subunits,
                step=step,
                system_potential=system_potential,
                nonbonded_potential=nonbonded_potential,
            )
            system_potential = step_result.system_potential
            nonbonded_potential = step_result.nonbonded_potential

        return mol, step_result



"""Atom class."""

from __future__ import annotations

from dataclasses import dataclass

from .radii import get_radius


@dataclass
class Atom:
    """Atom.

    Parameters:
        id:
            ID to be assigned to atom.

        element_string:
            Atom element symbol as string.

        radius:
            Radius (Default is from STREUSSEL) in angstrom.

        sigma:
            Value of sigma for custom potentials.

        epsilon:
            Value of epsilon for custom potentials.

        charge:
            Value of atomic charge for custom potentials.

    """

    id: int
    element_string: str
    radius: float | None = None
    sigma: float | None = None
    epsilon: float | None = None
    charge: float | None = None

    def __post_init__(self) -> None:
        """Post initialization of atom."""
        if self.radius is None:
            self.radius = get_radius(self.element_string)

    def get_id(self) -> int:
        """Get atom ID."""
        return self.id

    def get_element_string(self) -> str:
        """Get atom element symbol."""
        return self.element_string

    def get_radius(self) -> float | None:
        """Get atomic radius (STREUSEL)."""
        return self.radius

    def __str__(self) -> str:
        """String representation of Atom."""
        return repr(self)

    def __repr__(self) -> str:
        """String representation of Atom."""
        return f"{self.get_element_string()}(id={self.get_id()})"
